# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Handling of GitLab keep-arounds.

GitLab uses Git refs to prevent garbage collecting of commits that are
referenced in its database. While Mercurial does not have a garbage collector
it could grow one in the future for obsolete changesets, hence we better
not lose that information.

There is no reason to store keep-arounds as typed refs in Mercurial context,
though. We use a simpler format.

File format:

- version preamble (same as for typed refs)
- one hexadecimal Mercurial changeset ID per line.

Writes are append-only, there is no update or removal of keep-arounds.
For reads, we provide only a simple iterator.
"""
from contextlib import contextmanager
from .typed_ref import (
    gitlab_typed_refs_file_name,
    read_gitlab_typed_refs_file_version,
    write_gitlab_typed_refs_file_version,
    GITLAB_TYPED_REFS_MISSING,
)
TYPED_REF_NAME = 'keep-arounds'

KEEP_AROUND_REF_PREFIX = b'refs/keep-around/'
KEEP_AROUND_REF_PREFIX_LEN = len(KEEP_AROUND_REF_PREFIX)


def parse_keep_around_ref(ref_path):
    """Return target sha from ref_path, or ``None`` if not a keep-around ref.

    :param ref_path: full ref path, as :class:`bytes`
    :rtype: :class:`bytes

    Examples::

      >>> parse_keep_around_ref(
      ...     b'refs/keep-around/006e626febb896ca416bec68734a6199f4de5632')
      b'006e626febb896ca416bec68734a6199f4de5632'
      >>> parse_keep_around_ref(b'refs/merge-requests/1/head') is None
      True

    Special case for robustness, uniformity with other refs and convenience::

      >>> parse_keep_around_ref(None) is None
      True
    """
    if ref_path is None or not ref_path.startswith(KEEP_AROUND_REF_PREFIX):
        return None
    return ref_path[KEEP_AROUND_REF_PREFIX_LEN:]


@contextmanager
def open_keep_arounds_file(repo):
    """Context manager that opens the keep-arounds state file.

    The file format version is checked and the file handle is left after
    the preamble.

    If the state file is missing, yields :const:`GITLAB_TYPED_REFS_MISSING`
    """
    try:
        with repo.svfs(gitlab_typed_refs_file_name(TYPED_REF_NAME),
                       mode=b'rb') as fobj:
            assert read_gitlab_typed_refs_file_version(fobj) == 1
            yield fobj
    except FileNotFoundError:
        yield GITLAB_TYPED_REFS_MISSING


def iter_keep_arounds(repo):
    """Generator for existing keep-arounds.

    If the state file is missing, yields only
    :const:`GITLAB_TYPED_REFS_MISSING`.

    A future garbage collector could use this method to create a set of
    changesets to exclude from collection.
    """
    with open_keep_arounds_file(repo) as fobj:
        if fobj is GITLAB_TYPED_REFS_MISSING:
            yield fobj
            return

        for line in fobj:
            yield line.strip()


def create_keep_around(repo, hg_sha):
    """Write a new keep around to disk.

    :param hg_sha: Mercurial hexadecimal changeset ID, given as :class:`bytes`
    """
    with repo.lock():
        with repo.svfs(gitlab_typed_refs_file_name(TYPED_REF_NAME),
                       mode=b'ab',
                       checkambig=True) as fobj:
            if fobj.tell() == 0:  # new file
                write_gitlab_typed_refs_file_version(fobj)
            fobj.write(hg_sha)
            fobj.write(b'\n')


def init_keep_arounds(repo, hg_shas):
    """Initialize the state file from any iterable of SHAs.

    The state file is created or entirely remplaced.
    :param hg_shas: iterable of Mercurial hexadecimal changeset IDs,
        given as :class:`bytes`
    """
    with repo.lock():
        with repo.svfs(gitlab_typed_refs_file_name(TYPED_REF_NAME),
                       mode=b'wb',
                       atomictemp=True,
                       checkambig=True) as fobj:
            write_gitlab_typed_refs_file_version(fobj)
            for hg_sha in hg_shas:
                fobj.write(hg_sha)
                fobj.write(b'\n')



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
import pytest
from mercurial import (
    error,
)
from heptapod.testhelpers import (
    RepoWrapper,
    make_ui,
    as_bytes,
)
from ..topic import (
    change_topic,
    cleanup_tmp_bundle,
)
from .utils import common_config


def make_main_repo(path):
    """Make repo with 2 public revs; return wrapper, ctx of rev 0

    The returned ctx is for the first changeset because we'll use it as
    a branching point, hence more often than the second.
    """
    config = common_config()
    config['phases'] = dict(publish=False)
    wrapper = RepoWrapper.init(path, config=config)

    ctx = wrapper.write_commit('foo', content='foo0', message="default0",
                               return_ctx=True)
    wrapper.write_commit('foo', content='foo1', message="default1")
    wrapper.set_phase('public', ['.'])
    return wrapper, ctx


def test_force_draft(tmpdir):
    """Case where the remote changesets are public."""
    main_path = tmpdir.join('main')
    fork_path = tmpdir.join('fork')

    main, base_ctx = make_main_repo(main_path)

    fork = RepoWrapper.init(fork_path)
    fork.command('pull', as_bytes(main_path), rev=[base_ctx.hex()])
    fork_ctx = fork.write_commit('bar', content='bar1', message="in fork 1",
                                 parent=base_ctx)
    fork.set_phase('public', ['.'])

    code = main.command('pull-force-topic', b'zetop',
                        source=as_bytes(fork_path), rev=[fork_ctx.hex()])
    assert code == 0

    # TODO better lookup ?
    imported = main.repo[b'tip']
    assert imported.topic() == b'zetop'  # implied draft
    assert imported.description() == b'in fork 1'


def test_empty_pull(tmpdir):
    """Case where the pull turns out to be empty.

    See also heptapod#169
    """
    main_path = tmpdir.join('main')
    fork_path = tmpdir.join('fork')

    main, base_ctx = make_main_repo(main_path)

    fork = RepoWrapper.init(fork_path)
    fork.command('pull', as_bytes(main_path), rev=[base_ctx.hex()])
    fork_ctx = fork.write_commit('bar', content='bar1', message="in fork 1",
                                 parent=base_ctx)
    fork.set_phase('public', ['.'])

    main.command('pull', source=as_bytes(fork_path))
    code = main.command('pull-force-topic', b'zetop',
                        source=as_bytes(fork_path), rev=[fork_ctx.hex()])
    assert code == 1


def test_change_topic_sanity(tmpdir):
    wrapper, base_ctx = make_main_repo(tmpdir)
    repo = wrapper.repo

    # early validation:
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('pull-force-topic', b' ')
    assert b'entirely of whitespace' in exc_info.value.args[0]

    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('pull-force-topic', b'tip')
    assert b'reserved' in exc_info.value.args[0]

    # inner check done in transaction
    with pytest.raises(error.Abort) as exc_info:
        change_topic(repo.ui, repo, b'sometopic', [base_ctx.node()])
    assert b'public' in exc_info.value.args[0]


def test_cleanup_tmp_bundle_exc(tmpdir):
    ui = make_ui(None)

    path = tmpdir.join('bundle')

    # no file at given path
    cleanup_tmp_bundle(ui, None, as_bytes(path))

    path.write('foo')
    # close fails (of course, fobj is None)
    cleanup_tmp_bundle(ui, None, as_bytes(path))


def test_already_obsolete(tmpdir):
    """Case where a remote changeset is locally obsolete

    This can happen in practice if two remote pulls are stacked and
    we already pulled one, making it actually obsolete because of the
    topic forcing.
    """
    main_path = tmpdir.join('main')
    fork_path = tmpdir.join('fork')

    main = make_main_repo(main_path)[0]
    amended = main.write_commit('foo', content='foo2', return_ctx=True)

    fork = RepoWrapper.init(fork_path)
    fork.command('pull', as_bytes(main_path), rev=[amended.hex()])

    main_path.join('foo').write('amended')
    main.command('amend', message=b"required in CI environment")

    main.command('pull-force-topic', b'zetop',
                 source=as_bytes(fork_path), rev=[amended.hex()])


def test_overlapping(tmpdir):
    """Case where two intersection pull_force_topic are called.

    this is heptapod#226, with at changeset part of several PRs
    """
    main_path = tmpdir.join('main')
    # actually we could use just one source, but it would be
    # less clear at first sight
    fork1_path = tmpdir.join('fork1')
    fork2_path = tmpdir.join('fork2')

    main, base_ctx = make_main_repo(main_path)

    fork1 = RepoWrapper.init(fork1_path)
    fork1.command('pull', as_bytes(main_path), rev=[base_ctx.hex()])
    fork1_ctx = fork1.write_commit('bar', content='bar1', message="in fork 1",
                                   parent=base_ctx)
    fork1.set_phase('public', ['.'])

    fork2 = RepoWrapper.init(fork2_path)
    fork2.command('pull', as_bytes(fork1_path), rev=[fork1_ctx.hex()])
    fork2_ctx = fork2.write_commit('bar', content='bar2', message="in fork 2",
                                   branch='other',
                                   parent=fork1_ctx)
    fork2.set_phase('public', ['.'])

    code = main.command('pull-force-topic', b'zetop',
                        source=as_bytes(fork1_path),
                        rev=[fork1_ctx.hex()])
    assert code == 0

    imported1 = main.repo[b'tip']
    assert imported1.topic() == b'zetop'  # implied draft
    assert imported1.description() == b'in fork 1'

    code = main.command('pull-force-topic', b'othertop',
                        source=as_bytes(fork2_path),
                        rev=[fork2_ctx.hex()])
    assert code == 0

    # TODO better lookup ?
    imported2 = main.repo[b'tip']
    assert imported2.topic() == b'othertop'  # implied draft
    assert imported2.description() == b'in fork 2'

    assert not imported1.instabilities()
    assert not {b'content-divergent', b'phase-divergent'}.intersection(
        imported2.instabilities())


def test_non_obsolete_revs(tmpdir):
    repo_path = tmpdir.join('repo')
    wrapper, base_ctx = make_main_repo(repo_path)
    amended = wrapper.write_commit('foo', content='foo2', return_ctx=False)

    repo_path.join('foo').write('amended')
    wrapper.command('amend', message=b"required in CI environment")

    from hgext3rd.heptapod.topic import non_obsolete_revs
    assert list(non_obsolete_revs(wrapper.repo,
                                  [base_ctx.node(), amended])
                ) == [0]



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Compatibility layer for externalisation of testhelpers."""

from contextlib import contextmanager
import os

from mercurial import (
    util,
)
from mercurial_testhelpers.repo_wrapper import (
    RepoWrapper as CoreRepoWrapper,
    as_bytes,
)


class RepoWrapper(CoreRepoWrapper):

    write_commit = CoreRepoWrapper.commit_file
    commit_remove = CoreRepoWrapper.commit_removal

    def prune(self, revs, successors=(), bookmarks=()):
        # the prune command expects to get all these arguments (it relies
        # on the CLI defaults but doesn't have any at the function call level).
        # They are unconditionally concatened to lists, hence must be lists.
        # (as of Mercurial 5.3.1)
        if isinstance(revs, (bytes, str)):
            revs = [revs]
        return self.command('prune',
                            rev=[as_bytes(r) for r in revs],
                            new=[],  # deprecated yet expected
                            # TODO py3 convert these two to bytes as needed:
                            successor=list(successors),
                            bookmark=list(bookmarks))

    def set_topic(self, topic):
        self.command('topics', as_bytes(topic))

    def amend_file(self, relative_path, content=None, message=None):
        if content is None:
            content = self.random_content()
        content = as_bytes(content)

        if message is None:
            message = content
        else:
            message = as_bytes(message)

        path = os.path.join(self.repo.root, as_bytes(relative_path))
        with open(path, 'wb') as fobj:
            fobj.write(content)
        self.command('amend', message=message)
        return self.repo[b'tip']

    def clear_tags_cache(self):
        """Reset the repo's tag cache.

        This can be necessary in testing scenarios where several
        tags are created and inspected. Otherwise all calls to ``repo.tags()``
        would return the same value, even if more tags have been created
        since the first one.
        """
        # in actual scenario for StateMaintainer, the unfiltered is the
        # one that matters
        for repo in (self.repo, self.repo.unfiltered()):
            util.clearcachedproperty(repo, '_tagscache')

    @contextmanager
    def tags_sanitizer(self):
        """Restore repo tags to usable state after running test code.

        The tags caches must be reset in order not to pollute subsequent usage.
        Some failure remnants must be reverted.
        """
        self.clear_tags_cache()

        yield

        self.clear_tags_cache()
        # aborted transactions can leave local mods in .hgtags
        self.command('revert', as_bytes(self.path / '.hgtags'))


RepoWrapper.register_commit_option('topic', 'set_topic')

# compatibility alias
LocalRepoWrapper = RepoWrapper



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
from os.path import (
    dirname,
    join,
)
import heptapod
import hgext3rd.heptapod
HGEXT_HEPTA_SOURCE = dirname(hgext3rd.heptapod.__file__)

HEPTAPOD_REQUIRED_HGRC = join(os.fsencode(dirname(heptapod.__file__)),
                              b'required.hgrc')


def common_config():
    """Return a configuration dict, prefilled with what we almost always need.

    Notably, it activates the present extension.
    """
    return dict(extensions=dict(heptapod=HGEXT_HEPTA_SOURCE,
                                topic='',
                                evolve='',
                                rebase='',
                                ))



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os

import dulwich
from dulwich.object_store import (
    DiskObjectStore,
    INFODIR,
    GitFile,
)


# backport from dulwich commits:
# 06b65c3c2aef57939143e33d404f9883ec286597
# 544e90f8cbce1282ecb3d8ff73b5d7ecae905601
#
# see also https://github.com/dulwich/dulwich/pull/815 and heptapod#370
def _read_alternate_paths(self):  # pragma: no cover
    try:
        f = GitFile(os.path.join(self.path, INFODIR, "alternates"), 'rb')
    except FileNotFoundError:
        return
    with f:
        for line in f.readlines():
            line = line.rstrip(b"\n")
            if line.startswith(b"#"):
                continue
            if os.path.isabs(line):
                yield os.fsdecode(line)
            else:
                yield os.fsdecode(os.path.join(os.fsencode(self.path),
                                               line))


if dulwich.__version__ < (0, 20, 14):  # pragma no cover
    # our tests will run immediately on the dulwich with the fix,
    # only coverage telling us it's time to remove the unneeded patch
    DiskObjectStore._read_alternate_paths = _read_alternate_paths



# Copyright 2020-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Heptapod facilities for GitLab tags.

GitLab tags are a case of :mod:`typed refs <hgext3rd.heptapod.typed_ref>`,
we store them in their own state file, which represents the current view
by the other GitLab components, hence if there is an ongoing transaction,
it acts as a snapshot of the state at the beginning of the
transaction, so that we can notify them about the changes at the end
of the transaction.

We would not need this if Mercurial had a way to access tags changes from
a transaction object. Maybe it will in the future.
"""
from .typed_ref import (
    gitlab_typed_refs,
    read_gitlab_typed_refs,
    write_gitlab_typed_refs,
    remove_gitlab_typed_refs,
)


def read_gitlab_tags(repo):
    return read_gitlab_typed_refs(repo, 'tags')


def gitlab_tags(repo):
    return gitlab_typed_refs(repo, 'tags')


def write_gitlab_tags(repo, gl_tags):
    """Write the GitLab tags mapping atomically.
    """
    return write_gitlab_typed_refs(repo, 'tags', gl_tags)


def remove_gitlab_tags(repo):
    remove_gitlab_typed_refs(repo, 'tags')



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Test support for users of heptapod.gitlab
"""
from __future__ import absolute_import
import attr
from copy import deepcopy
import logging
import shutil

from heptapod.gitlab import hooks

from .hg import RepoWrapper
from .git import GitRepo
logger = logging.getLogger(__name__)


def patch_gitlab_hooks(monkeypatch, records, action=None):

    def call(self, changes):
        records.append((self.name, changes))
        if action is not None:
            return action(self.name, changes)
        else:
            return 0, ("hook %r ok" % self.name).encode(), 'no error'

    def init(self, repo, encoding='utf-8'):
        self.repo = repo
        self.encoding = encoding

    monkeypatch.setattr(hooks.Hook, '__init__', init)
    monkeypatch.setattr(hooks.PreReceive, '__call__', call)
    monkeypatch.setattr(hooks.PostReceive, '__call__', call)


@attr.s
class GitLabMirrorFixture:
    """Helper class to create fixtures for GitLab aware hg-git mirroring.

    The pytest fixture functions themselves will have to be provided with
    the tests that use them.

    It is not the role of this class to make decisions about scopes or
    the kind of root directory it operates in.

    There will be later on a fixture for Mercurial-native Heptapod
    repositories, hence GitLab notifications without a Git repository.
    """
    base_path = attr.ib()
    hg_repo_wrapper = attr.ib()
    git_repo = attr.ib()
    gitlab_notifs = attr.ib()
    import heptapod.testhelpers.gitlab

    @classmethod
    def init(cls, base_path, monkeypatch, hg_config=None,
             common_repo_name='repo'):
        if hg_config is None:
            config = {}
        else:
            config = deepcopy(hg_config)

        config.setdefault('extensions', {})['hggit'] = ''
        config['phases'] = dict(publish=False)

        hg_repo_wrapper = RepoWrapper.init(
            base_path / (common_repo_name + '.hg'),
            config=config)
        git_repo = GitRepo.init(base_path / (common_repo_name + '.git'))
        notifs = []
        patch_gitlab_hooks(monkeypatch, notifs)
        return cls(hg_repo_wrapper=hg_repo_wrapper,
                   git_repo=git_repo,
                   gitlab_notifs=notifs,
                   base_path=base_path)

    def clear_gitlab_notifs(self):
        """Forget about all notifications already sent to GitLab.

        Subsequent notifications will keep on being recorded in
        ``self.gitlab_notifs``.
        """
        del self.gitlab_notifs[:]

    def activate_mirror(self):
        """Make mirroring from Mercurial to Git repo automatic.

        This is essential to get the mirroring code to run in-transaction.
        """
        self.hg_repo_wrapper.repo.ui.setconfig(
            b'hooks', b'pretxnclose.testcase',
            b'python:heptapod.hooks.gitlab_mirror.mirror')

    def delete(self):
        git_path = self.git_repo.path
        try:
            shutil.rmtree(git_path)
        except Exception:
            logger.exception("Error removing the Git repo at %r", git_path)

        hg_path = self.hg_repo_wrapper.repo.root
        try:
            shutil.rmtree(hg_path)
        except Exception:
            logger.exception("Error removing the Mercurial repo at %r",
                             hg_path)

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        self.delete()
        return False  # no exception handling related to exc_args



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
import tarfile
from io import BytesIO

from mercurial import (
    error,
    scmutil,
)
from heptapod.testhelpers import (
    RepoWrapper,
    as_bytes,
)
from .utils import common_config

from .. import (
    backup_additional,
    restore_additional,
)
from ..branch import (
    gitlab_branches,
    get_default_gitlab_branch,
)
from ..git import HeptapodGitHandler
from ..keep_around import (
    create_keep_around,
    iter_keep_arounds,
)
from ..special_ref import (
    special_refs,
    write_gitlab_special_ref,
)
from ..tag import gitlab_tags

from heptapod.testhelpers.gitlab import GitLabMirrorFixture


def test_additional_backup_restore(tmpdir, monkeypatch):
    fixture = GitLabMirrorFixture.init(tmpdir, monkeypatch,
                                       hg_config=common_config())
    src_wrapper = fixture.hg_repo_wrapper
    src_repo = src_wrapper.repo
    main_hgrc = b'[foo]\nsome=bar\n'
    managed_hgrc = b'[heptapod]\nallow-bookmars=true\n'
    src_repo.vfs.write(b'hgrc', main_hgrc)
    src_repo.vfs.write(b'hgrc.managed', managed_hgrc)

    # Backup/restore for the empty repo to check that config
    # is still ok (even though default branch file is missing)
    tar_for_empty_path = tmpdir / 'additional_empty_repo.tar'
    src_wrapper.command('hpd-backup-additional', as_bytes(tar_for_empty_path))

    empty_dst_wrapper = RepoWrapper.init(tmpdir / 'empty_restored')
    empty_dst_wrapper.command('hpd-restore-additional',
                              as_bytes(tar_for_empty_path))
    empty_dst_repo = empty_dst_wrapper.repo
    assert empty_dst_repo.vfs(b'hgrc').read() == main_hgrc
    assert empty_dst_repo.vfs(b'hgrc.managed').read() == managed_hgrc

    # Now with repo content and GitLab refs
    sha = src_wrapper.write_commit('foo', branch='surprise').hex()
    src_wrapper.command('tag', b'v3.2.1')
    tagging_sha = scmutil.revsingle(src_wrapper.repo, b'.').hex()

    src_wrapper.command('gitlab-mirror')
    write_gitlab_special_ref(src_wrapper.repo, b'merge_requests/5/head', sha)
    create_keep_around(src_repo, sha)

    tar_path = as_bytes(tmpdir / 'additional.tar')
    backup_additional(src_repo.ui, src_repo, tar_path)

    dst_wrapper = RepoWrapper.init(tmpdir / 'restored')
    dst_repo = dst_wrapper.repo
    dst_wrapper.command('pull', src_repo.root)

    restore_additional(dst_repo.ui, dst_repo, tar_path)

    # Default GitLab branch (cannot pass by accident thanks to branch name)
    # Assertions for all kinds of GitLab refs
    assert get_default_gitlab_branch(src_repo) == b'branch/surprise'
    assert special_refs(dst_repo) == {b'merge_requests/5/head': sha}
    assert list(iter_keep_arounds(dst_repo)) == [sha]
    assert gitlab_branches(dst_repo) == {b'branch/surprise': tagging_sha}
    assert gitlab_tags(dst_repo) == {b'v3.2.1': sha}

    # Assertions for hg-git mapping
    src_handler = HeptapodGitHandler(src_repo, src_repo.ui)
    dst_handler = HeptapodGitHandler(dst_repo, dst_repo.ui)
    dst_git_sha = dst_handler.map_git_get(sha)
    assert dst_git_sha  # avoid traps
    assert dst_git_sha == src_handler.map_git_get(sha)


def test_restore_bad_tar(tmpdir):
    dst_wrapper = RepoWrapper.init(tmpdir / 'repo')
    tar_path = tmpdir / 'bad.tar'

    # str() because we support Python 3.7 (for CI)
    # and path-like objects are accepted in tarfile from Python 3.8 onwards
    with tarfile.open(str(tar_path), 'w') as tarf:
        tinfo = tarfile.TarInfo(name='/etc/shadow')
        tarf.addfile(tinfo, BytesIO(b'evil input'))

    with pytest.raises(error.Abort) as exc_info:
        dst_wrapper.command('hpd-restore-additional', as_bytes(tar_path))

    assert b"Prohibited member" in exc_info.value.message



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from mercurial import (
    error,
)
from .perm import validate_hook_type


def forbid_commit(repo, *args, **kwargs):
    validate_hook_type(b'precommit', **kwargs)
    if repo.wvfs.exists(b'.hgsub'):
        raise error.Abort(b"This version of Heptapod cannot commit "
                          b"server-side if there are subrepos around. "
                          b"That could ruin .hgsubstate. "
                          b"A subsequent version may reenable this.")
